from abc import abstractmethod

import math

import numpy as np
import torch as th
import torch.nn as nn
import torch.nn.functional as F

from .fp16_util import convert_module_to_f16, convert_module_to_f32
from .nn import (
    checkpoint,
    conv_nd,
    linear,
    avg_pool_nd,
    zero_module,
    normalization,
    timestep_embedding,
)
from . import logger

class Pointer:
    def __init__(self):
        self.p = 0

pointer = Pointer()


class AttentionPool2d(nn.Module):
    """
    Adapted from CLIP: https://github.com/openai/CLIP/blob/main/clip/model.py
    """

    def __init__(
        self,
        spacial_dim: int,
        embed_dim: int,
        num_heads_channels: int,
        output_dim: int = None,
    ):
        super().__init__()
        self.positional_embedding = nn.Parameter(
            th.randn(embed_dim, spacial_dim ** 2 + 1) / embed_dim ** 0.5
        )
        self.qkv_proj = conv_nd(1, embed_dim, 3 * embed_dim, 1)
        self.c_proj = conv_nd(1, embed_dim, output_dim or embed_dim, 1)
        self.num_heads = embed_dim // num_heads_channels
        self.attention = QKVAttention(self.num_heads)

    def forward(self, x):
        b, c, *_spatial = x.shape
        x = x.reshape(b, c, -1)  # NC(HW)
        x = th.cat([x.mean(dim=-1, keepdim=True), x], dim=-1)  # NC(HW+1)
        x = x + self.positional_embedding[None, :, :].to(x.dtype)  # NC(HW+1)
        x = self.qkv_proj(x)
        x = self.attention(x)
        x = self.c_proj(x)
        return x[:, :, 0]


class TimestepBlock(nn.Module):
    """
    Any module where forward() takes timestep embeddings as a second argument.
    """

    @abstractmethod
    def forward(self, x, emb):
        """
        Apply the module to `x` given `emb` timestep embeddings.
        """


class TimestepEmbedSequential(nn.Sequential, TimestepBlock):
    """
    A sequential module that passes timestep embeddings to the children that
    support it as an extra input.
    """

    def forward(self, x, emb):
        for layer in self:
            if isinstance(layer, TimestepBlock):
                x = layer(x, emb)
            else:
                x = layer(x)
        return x

    def my_sample(self, x, emb):
        for layer in self:
            if isinstance(layer, TimestepBlock):
                x = layer.my_sample(x, emb)
            else:
                x = layer(x)
        return x


class Upsample(nn.Module):
    """
    An upsampling layer with an optional convolution.

    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 upsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, dims=2, out_channels=None):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.dims = dims
        if use_conv:
            self.conv = conv_nd(dims, self.channels, self.out_channels, 3, padding=1)

    def forward(self, x):
        assert x.shape[1] == self.channels
        if self.dims == 3:
            x = F.interpolate(
                x, (x.shape[2], x.shape[3] * 2, x.shape[4] * 2), mode="nearest"
            )
        else:
            x = F.interpolate(x, scale_factor=2, mode="nearest")
        if self.use_conv:
            x = self.conv(x)
        return x


class Downsample(nn.Module):
    """
    A downsampling layer with an optional convolution.

    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 downsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, dims=2, out_channels=None):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.dims = dims
        stride = 2 if dims != 3 else (1, 2, 2)
        if use_conv:
            self.op = conv_nd(
                dims, self.channels, self.out_channels, 3, stride=stride, padding=1
            )
        else:
            assert self.channels == self.out_channels
            self.op = avg_pool_nd(dims, kernel_size=stride, stride=stride)

    def forward(self, x):
        assert x.shape[1] == self.channels
        return self.op(x)


class ResBlock(TimestepBlock):
    """
    A residual block that can optionally change the number of channels.

    :param channels: the number of input channels.
    :param emb_channels: the number of timestep embedding channels.
    :param dropout: the rate of dropout.
    :param out_channels: if specified, the number of out channels.
    :param use_conv: if True and out_channels is specified, use a spatial
        convolution instead of a smaller 1x1 convolution to change the
        channels in the skip connection.
    :param dims: determines if the signal is 1D, 2D, or 3D.
    :param use_checkpoint: if True, use gradient checkpointing on this module.
    :param up: if True, use this block for upsampling.
    :param down: if True, use this block for downsampling.
    """

    def __init__(
        self,
        channels,
        emb_channels,
        dropout,
        out_channels=None,
        use_conv=False,
        use_scale_shift_norm=False,
        dims=2,
        use_checkpoint=False,
        up=False,
        down=False,
    ):
        super().__init__()
        self.channels = channels
        self.emb_channels = emb_channels
        self.dropout = dropout
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_checkpoint = use_checkpoint
        self.use_scale_shift_norm = use_scale_shift_norm

        self.in_layers = nn.Sequential(
            normalization(channels),
            nn.SiLU(),
            conv_nd(dims, channels, self.out_channels, 3, padding=1),
        )

        self.updown = up or down

        if up:
            self.h_upd = Upsample(channels, False, dims)
            self.x_upd = Upsample(channels, False, dims)
        elif down:
            self.h_upd = Downsample(channels, False, dims)
            self.x_upd = Downsample(channels, False, dims)
        else:
            self.h_upd = self.x_upd = nn.Identity()

        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            linear(
                emb_channels,
                2 * self.out_channels if use_scale_shift_norm else self.out_channels,
            ),
        )
        self.out_layers = nn.Sequential(
            normalization(self.out_channels),
            nn.SiLU(),
            nn.Dropout(p=dropout),
            zero_module(
                conv_nd(dims, self.out_channels, self.out_channels, 3, padding=1)
            ),
        )

        if self.out_channels == channels:
            self.skip_connection = nn.Identity()
        elif use_conv:
            self.skip_connection = conv_nd(
                dims, channels, self.out_channels, 3, padding=1
            )
        else:
            self.skip_connection = conv_nd(dims, channels, self.out_channels, 1)

    def forward(self, x, emb):
        """
        Apply the block to a Tensor, conditioned on a timestep embedding.

        :param x: an [N x C x ...] Tensor of features.
        :param emb: an [N x emb_channels] Tensor of timestep embeddings.
        :return: an [N x C x ...] Tensor of outputs.
        """
        return checkpoint(
            self._forward, (x, emb), self.parameters(), self.use_checkpoint
        )

    def _forward(self, x, emb):
        if self.updown:
            in_rest, in_conv = self.in_layers[:-1], self.in_layers[-1]
            h = in_rest(x)
            h = self.h_upd(h)
            x = self.x_upd(x)
            h = in_conv(h)
        else:
            h = self.in_layers(x)
        emb_out = self.emb_layers(emb).type(h.dtype)
        while len(emb_out.shape) < len(h.shape):
            emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            out_norm, out_rest = self.out_layers[0], self.out_layers[1:]
            scale, shift = th.chunk(emb_out, 2, dim=1)
            h = out_norm(h) * (1 + scale) + shift
            h = out_rest(h)
        else:
            h = h + emb_out
            h = self.out_layers(h)
        return self.skip_connection(x) + h


class AttentionBlock(nn.Module):
    """
    An attention block that allows spatial positions to attend to each other.

    Originally ported from here, but adapted to the N-d case.
    https://github.com/hojonathanho/diffusion/blob/1e0dceb3b3495bbe19116a5e1b3596cd0706c543/diffusion_tf/models/unet.py#L66.
    """

    def __init__(
        self,
        channels,
        num_heads=1,
        num_head_channels=-1,
        use_checkpoint=False,
        use_new_attention_order=False,
    ):
        super().__init__()
        self.channels = channels
        if num_head_channels == -1:
            self.num_heads = num_heads
        else:
            assert (
                channels % num_head_channels == 0
            ), f"q,k,v channels {channels} is not divisible by num_head_channels {num_head_channels}"
            self.num_heads = channels // num_head_channels
        self.use_checkpoint = use_checkpoint
        self.norm = normalization(channels)
        self.qkv = conv_nd(1, channels, channels * 3, 1)
        if use_new_attention_order:
            # split qkv before split heads
            self.attention = QKVAttention(self.num_heads)
        else:
            # split heads before split qkv
            self.attention = QKVAttentionLegacy(self.num_heads)

        self.proj_out = zero_module(conv_nd(1, channels, channels, 1))

    def forward(self, x):
        return checkpoint(self._forward, (x,), self.parameters(), True)

    def _forward(self, x):
        b, c, *spatial = x.shape
        x = x.reshape(b, c, -1)
        qkv = self.qkv(self.norm(x))
        h = self.attention(qkv)
        h = self.proj_out(h)
        return (x + h).reshape(b, c, *spatial)


def count_flops_attn(model, _x, y):
    """
    A counter for the `thop` package to count the operations in an
    attention operation.
    Meant to be used like:
        macs, params = thop.profile(
            model,
            inputs=(inputs, timestamps),
            custom_ops={QKVAttention: QKVAttention.count_flops},
        )
    """
    b, c, *spatial = y[0].shape
    num_spatial = int(np.prod(spatial))
    # We perform two matmuls with the same number of ops.
    # The first computes the weight matrix, the second computes
    # the combination of the value vectors.
    matmul_ops = 2 * b * (num_spatial ** 2) * c
    model.total_ops += th.DoubleTensor([matmul_ops])


class QKVAttentionLegacy(nn.Module):
    """
    A module which performs QKV attention. Matches legacy QKVAttention + input/ouput heads shaping
    """

    def __init__(self, n_heads):
        super().__init__()
        self.n_heads = n_heads

    def forward(self, qkv):
        """
        Apply QKV attention.

        :param qkv: an [N x (H * 3 * C) x T] tensor of Qs, Ks, and Vs.
        :return: an [N x (H * C) x T] tensor after attention.
        """
        bs, width, length = qkv.shape
        assert width % (3 * self.n_heads) == 0
        ch = width // (3 * self.n_heads)
        q, k, v = qkv.reshape(bs * self.n_heads, ch * 3, length).split(ch, dim=1)
        scale = 1 / math.sqrt(math.sqrt(ch))
        weight = th.einsum(
            "bct,bcs->bts", q * scale, k * scale
        )  # More stable with f16 than dividing afterwards
        weight = th.softmax(weight.float(), dim=-1).type(weight.dtype)
        a = th.einsum("bts,bcs->bct", weight, v)
        return a.reshape(bs, -1, length)

    @staticmethod
    def count_flops(model, _x, y):
        return count_flops_attn(model, _x, y)


class QKVAttention(nn.Module):
    """
    A module which performs QKV attention and splits in a different order.
    """

    def __init__(self, n_heads):
        super().__init__()
        self.n_heads = n_heads

    def forward(self, qkv):
        """
        Apply QKV attention.

        :param qkv: an [N x (3 * H * C) x T] tensor of Qs, Ks, and Vs.
        :return: an [N x (H * C) x T] tensor after attention.
        """
        bs, width, length = qkv.shape
        assert width % (3 * self.n_heads) == 0
        ch = width // (3 * self.n_heads)
        q, k, v = qkv.chunk(3, dim=1)
        scale = 1 / math.sqrt(math.sqrt(ch))
        weight = th.einsum(
            "bct,bcs->bts",
            (q * scale).view(bs * self.n_heads, ch, length),
            (k * scale).view(bs * self.n_heads, ch, length),
        )  # More stable with f16 than dividing afterwards
        weight = th.softmax(weight.float(), dim=-1).type(weight.dtype)
        a = th.einsum("bts,bcs->bct", weight, v.reshape(bs * self.n_heads, ch, length))
        return a.reshape(bs, -1, length)

    @staticmethod
    def count_flops(model, _x, y):
        return count_flops_attn(model, _x, y)


class UNetModel(nn.Module):
    """
    The full UNet model with attention and timestep embedding.

    :param in_channels: channels in the input Tensor.
    :param model_channels: base channel count for the model.
    :param out_channels: channels in the output Tensor.
    :param num_res_blocks: number of residual blocks per downsample.
    :param attention_resolutions: a collection of downsample rates at which
        attention will take place. May be a set, list, or tuple.
        For example, if this contains 4, then at 4x downsampling, attention
        will be used.
    :param dropout: the dropout probability.
    :param channel_mult: channel multiplier for each level of the UNet.
    :param conv_resample: if True, use learned convolutions for upsampling and
        downsampling.
    :param dims: determines if the signal is 1D, 2D, or 3D.
    :param num_classes: if specified (as an int), then this model will be
        class-conditional with `num_classes` classes.
    :param use_checkpoint: use gradient checkpointing to reduce memory usage.
    :param num_heads: the number of attention heads in each attention layer.
    :param num_heads_channels: if specified, ignore num_heads and instead use
                               a fixed channel width per attention head.
    :param num_heads_upsample: works with num_heads to set a different number
                               of heads for upsampling. Deprecated.
    :param use_scale_shift_norm: use a FiLM-like conditioning mechanism.
    :param resblock_updown: use residual blocks for up/downsampling.
    :param use_new_attention_order: use a different attention pattern for potentially
                                    increased efficiency.
    """

    def __init__(
        self,
        image_size,
        in_channels,
        model_channels,
        out_channels,
        num_res_blocks,
        attention_resolutions,
        dropout=0,
        channel_mult=(1, 2, 4, 8),
        conv_resample=True,
        dims=2,
        num_classes=None,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=1,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=False,
        resblock_updown=False,
        use_new_attention_order=False,
        expand_scale=0.5,
        expand_scale_io=1,
        expand_interval=[3, 8, 11, 15],
        skip_1024layers=[11, 7],
        t_spli_num=10,
        timesteps=1000,
        threshold=0.49999,
        if_finetune=False,
        layer_used_num_list=[],
        double_mid=False,
    ):
        super().__init__()

        if num_heads_upsample == -1:
            num_heads_upsample = num_heads

        self.image_size = image_size
        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.attention_resolutions = attention_resolutions
        self.dropout = dropout
        self.channel_mult = channel_mult
        self.conv_resample = conv_resample
        self.num_classes = num_classes
        self.use_checkpoint = use_checkpoint
        self.dtype = th.float16 if use_fp16 else th.float32
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.num_heads_upsample = num_heads_upsample

        #################################################################
        # DDDM 的额外添加
        self.expand_scale = expand_scale
        # self.expand_scale_io = expand_scale_io
        self.gater_length = 0
        self.t_spli_num = t_spli_num
        self.steps = timesteps
        skip_in_1024layers, skip_out_1024layers = skip_1024layers[0], skip_1024layers[1]
        # exp_interval_in1, exp_interval_in2, exp_interval_out1, exp_interval_out2 = expand_interval[0], expand_interval[1], expand_interval[2], expand_interval[3]
        self.expand_scale_list = []
        self.if_finetune = if_finetune
        self.double_mid = double_mid

        if if_finetune:
            self.trained_scales = []
            for i in range(len(layer_used_num_list)):
                self.trained_scales.append(int(layer_used_num_list[i] * expand_scale))
                if self.trained_scales[i] == 0:
                    self.trained_scales[i] += 1

            self.init_p = 0
        #################################################################

        time_embed_dim = model_channels * 4
        self.time_embed = nn.Sequential(
            linear(model_channels, time_embed_dim),
            nn.SiLU(),
            linear(time_embed_dim, time_embed_dim),
        )

        if self.num_classes is not None:
            self.label_emb = nn.Embedding(num_classes, time_embed_dim)

        ch = input_ch = int(channel_mult[0] * model_channels)
        # self.input_blocks = nn.ModuleList(
        #     [TimestepEmbedSequential(conv_nd(dims, in_channels, ch, 3, padding=1))]
        # )
        if self.if_finetune:
            exp_scale = self.trained_scales[0]
        else:
            exp_scale = 1
        self.input_blocks = nn.ModuleList(
            [Input0(exp_scale, dims, in_channels, ch)]
        )
        self.expand_scale_list.append(exp_scale)
        logger.log("Inputblocks 0 is expanded by", exp_scale, "times in training process.")
        self.init_p += 1
        self.gater_length += exp_scale

        self._feature_size = ch
        input_block_chans = [ch]
        ds = 1
        for level, mult in enumerate(channel_mult):
            for _ in range(num_res_blocks):
                #########################
                if self.if_finetune:
                    exp_scale = self.trained_scales[self.init_p]
                else:
                    exp_scale = 1
                #########################
                layers = [
                    BigResBlock(
                        exp_scale,
                        ch,
                        time_embed_dim,
                        dropout,
                        out_channels=int(mult * model_channels),
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                if len(self.input_blocks) < skip_in_1024layers: #注意，这里的大块还没加到input_blocks里面，因此是严格小于
                    self.gater_length += exp_scale
                    logger.log("Inputblocks", len(self.input_blocks), "is expanded by", exp_scale, "times in training process.")
                    self.expand_scale_list.append(exp_scale)
                    self.init_p += 1

                ch = int(mult * model_channels)
                if ds in attention_resolutions:
                    #########################
                    if self.if_finetune:
                        exp_scale = self.trained_scales[self.init_p]
                    else:
                        exp_scale = 1
                    #########################
                    layers.append(
                        BigAttentionBlock(
                            exp_scale,
                            ch,
                            use_checkpoint=use_checkpoint,
                            num_heads=num_heads,
                            num_head_channels=num_head_channels,
                            use_new_attention_order=use_new_attention_order,
                        )
                    )
                    if len(self.input_blocks) < skip_in_1024layers:
                        self.gater_length += exp_scale
                        self.expand_scale_list.append(exp_scale)
                        logger.log("There is a Attention Block, it's expanded by", exp_scale, "times. Its index in gate vector is", len(self.expand_scale_list)-1)
                        self.init_p += 1

                self.input_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch
                input_block_chans.append(ch)
            if level != len(channel_mult) - 1:
                out_ch = ch
                #########################
                if self.if_finetune:
                    exp_scale = self.trained_scales[self.init_p]
                else:
                    exp_scale = 1
                #########################
                self.input_blocks.append(
                    TimestepEmbedSequential(
                        BigResBlock(
                            exp_scale,
                            ch,
                            time_embed_dim,
                            dropout,
                            out_channels=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            down=True,
                        )
                        if resblock_updown
                        else Downsample(
                            ch, conv_resample, dims=dims, out_channels=out_ch
                        )
                    )
                )
                if len(self.input_blocks) <= skip_in_1024layers: #注意，这里的大块已经加到input_blocks里面，因此是小于等于
                    self.gater_length += exp_scale
                    logger.log("Inputblocks", len(self.input_blocks)-1, "is expanded by", exp_scale, "times in training process.")
                    self.expand_scale_list.append(exp_scale)
                    self.init_p += 1

                ch = out_ch
                input_block_chans.append(ch)
                ds *= 2
                self._feature_size += ch

        self.middle_block = TimestepEmbedSequential(
            BigResBlock(
                1,
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
            AttentionBlock(
                ch,
                use_checkpoint=use_checkpoint,
                num_heads=num_heads,
                num_head_channels=num_head_channels,
                use_new_attention_order=use_new_attention_order,
            ),
            BigResBlock(
                1,
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
        )
        self._feature_size += ch

        self.output_blocks = nn.ModuleList([])
        for level, mult in list(enumerate(channel_mult))[::-1]:
            for i in range(num_res_blocks + 1):
                ich = input_block_chans.pop()
                #########################
                if self.if_finetune:
                    exp_scale = self.trained_scales[self.init_p]
                else:
                    exp_scale = 1
                #########################
                layers = [
                    BigResBlock(
                        exp_scale,
                        ch + ich,
                        time_embed_dim,
                        dropout,
                        out_channels=int(model_channels * mult),
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                if len(self.output_blocks) >= skip_out_1024layers: #注意，这里的大块还没加到output_blocks里面，因此是大于等于
                    self.gater_length += exp_scale
                    logger.log("Outputblocks", len(self.output_blocks), "is expanded by", exp_scale, "times in training process.")
                    self.expand_scale_list.append(exp_scale)
                    self.init_p += 1

                ch = int(model_channels * mult)
                if ds in attention_resolutions:
                    #########################
                    if self.if_finetune:
                        exp_scale = self.trained_scales[self.init_p]
                    else:
                        exp_scale = 1
                    #########################
                    layers.append(
                        BigAttentionBlock(
                            exp_scale,
                            ch,
                            use_checkpoint=use_checkpoint,
                            num_heads=num_heads_upsample,
                            num_head_channels=num_head_channels,
                            use_new_attention_order=use_new_attention_order,
                        )
                    )
                    if len(self.output_blocks) >= skip_out_1024layers:
                        self.gater_length += exp_scale
                        self.expand_scale_list.append(exp_scale)
                        logger.log("There is a Attention Block, it's expanded by", exp_scale, "times. Its index in gate vector is", len(self.expand_scale_list)-1)
                        self.init_p += 1

                if level and i == num_res_blocks:
                    out_ch = ch
                    #########################
                    if self.if_finetune:
                        exp_scale = self.trained_scales[self.init_p]
                    else:
                        exp_scale = 1
                    #########################
                    layers.append(
                        BigResBlock(
                            exp_scale,
                            ch,
                            time_embed_dim,
                            dropout,
                            out_channels=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            up=True,
                        )
                        if resblock_updown
                        else Upsample(ch, conv_resample, dims=dims, out_channels=out_ch)
                    )
                    if len(self.output_blocks) >= skip_out_1024layers: #注意，这里的大块还没加到output_blocks里面，因此是大于等于
                        self.gater_length += exp_scale
                        logger.log("Outputblocks", len(self.output_blocks), "is expanded by", exp_scale, "times in training process.")
                        self.expand_scale_list.append(exp_scale)
                        self.init_p += 1

                    ds //= 2
                self.output_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch

        # self.out = nn.Sequential(
        #     normalization(ch),
        #     nn.SiLU(),
        #     zero_module(conv_nd(dims, input_ch, out_channels, 3, padding=1)),
        # )
        #########################
        if self.if_finetune:
            exp_scale = self.trained_scales[self.init_p]
        else:
            exp_scale = 1
        #########################
        self.out = Out_layer(
            exp_scale, dims, ch, input_ch, out_channels
        )
        logger.log("The OutLayer is expanded by", exp_scale, "times in training process.")
        self.expand_scale_list.append(exp_scale)
        self.gater_length += exp_scale
        self.init_p += 1

        assert np.sum(self.expand_scale_list) == self.gater_length
        assert self.init_p == len(self.expand_scale_list)
        assert self.init_p == len(self.trained_scales)
        for i in range(self.init_p):
            assert self.expand_scale_list[i] == self.trained_scales[i]

        ####################################################
        # 直接跳过尺寸为1024的卷积Res-Blocks
        self.input_blocks = self.input_blocks[:skip_in_1024layers]
        self.output_blocks = self.output_blocks[skip_out_1024layers:]
        del self.middle_block

        if not self.if_finetune:
            self.gater_head = FNN(dims=[self.t_spli_num, 64, self.gater_length], threshold=threshold)

        self.threshold = threshold
        ####################################################

    def convert_to_fp16(self):
        """
        Convert the torso of the model to float16.
        """
        self.input_blocks.apply(convert_module_to_f16)
        # self.middle_block.apply(convert_module_to_f16)
        self.output_blocks.apply(convert_module_to_f16)

    def convert_to_fp32(self):
        """
        Convert the torso of the model to float32.
        """
        self.input_blocks.apply(convert_module_to_f32)
        # self.middle_block.apply(convert_module_to_f32)
        self.output_blocks.apply(convert_module_to_f32)

    def forward(self, x, timesteps, y=None, if_finetune=False, gates=None):
        """
        Apply the model to an input batch.

        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :param y: an [N] Tensor of labels, if class-conditional.
        :return: an [N x C x ...] Tensor of outputs.
        """
        assert (y is not None) == (
            self.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"

        hs = []
        emb = self.time_embed(timestep_embedding(timesteps, self.model_channels))

        if self.num_classes is not None:
            assert y.shape == (x.shape[0],)
            emb = emb + self.label_emb(y)

        ############################################
        # print(if_finetune, self.if_finetune)
        assert self.if_finetune == if_finetune
        feature_t = (timesteps * self.t_spli_num / self.steps).type(th.int64)
        one_hot_t = nn.functional.one_hot(feature_t, num_classes=self.t_spli_num)

        if not if_finetune:
            gates = self.gater_head(one_hot_t.float())
            gates = gates.type(self.dtype)

            pointer.p = 0
            ############################################

            h = x.type(self.dtype)
            for module in self.input_blocks:
                h = module(h, (emb, gates))
                hs.append(h)
            # 跳过尺寸为1024的卷积Res-Blocks
            # h = self.middle_block(h, emb)
            if self.double_mid:
                h = th.cat([h, h], dim=1)
            for module in self.output_blocks:
                h = th.cat([h, hs.pop()], dim=1)
                h = module(h, (emb, gates))
            #####
            #h = h.type(x.dtype)
            h = h.type(th.cuda.FloatTensor)
            out = self.out(h, (None, gates))
            assert pointer.p == gates.shape[1]
            out = out.type(x.dtype)
            if self.training:
                return out, gates
            return out

        if if_finetune:
            # gates = self.get_discrete_gate(timesteps)
            assert gates != None
            pointer.p = 0
            gates = gates.detach()
            h = x.type(self.dtype)
            gates = gates.type(self.dtype)
            for module in self.input_blocks:
                h = module(h, (emb, gates))
                hs.append(h)

            if self.double_mid:
                h = th.cat([h, h], dim=1)

            for module in self.output_blocks:
                h = th.cat([h, hs.pop()], dim=1)
                h = module(h, (emb, gates))

            h = h.type(th.cuda.FloatTensor)
            out = self.out(h, (None, gates))
            assert pointer.p == gates.shape[1]
            out = out.type(x.dtype)

            return out

    def get_discrete_gate(self, timesteps):
        feature_t = (timesteps * self.t_spli_num / self.steps).type(th.int64)
        one_hot_t = nn.functional.one_hot(feature_t, num_classes=self.t_spli_num)
        with th.no_grad():
            self.gater_head.eval()
            gates = self.gater_head(one_hot_t.float())
            len_gate = gates.shape[1]
            pointer.p = 0
            for exp_scale in self.expand_scale_list:
                start_id = pointer.p
                end_id = start_id + exp_scale
                pointer.p = end_id

                for i in range(gates.shape[0]):
                    scale_ = th.max(gates[i, start_id: end_id])
                    if scale_ > self.threshold:
                        scale = 1. / scale_
                    else:
                        scale = 0.

                    gates[i, start_id: end_id] *= scale
            
            assert pointer.p == len_gate

        gates = gates.int()
        io_expand_scale = self.expand_scale_list[0]
        for i in range(gates.shape[0]):
            if gates[i, 0:io_expand_scale].sum() == 0:
                t_scale = timesteps[i] / self.steps
                location = int(t_scale * io_expand_scale)
                gates[i, location] = 1
            if gates[i, -io_expand_scale:].sum() == 0:
                t_scale = timesteps[i] / self.steps
                location = int(t_scale * io_expand_scale)
                gates[i, -(io_expand_scale-location)] = 1
        return gates.int()

    def sample(self, x, timesteps, y=None, layer_idx=None):
        assert layer_idx != None
        pointer.p = 0

        assert (y is not None) == (
            self.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"

        hs = []
        emb = self.time_embed(timestep_embedding(timesteps, self.model_channels))

        if self.num_classes is not None:
            assert y.shape == (x.shape[0],)
            emb = emb + self.label_emb(y)

        h = x.type(self.dtype)
        for module in self.input_blocks:
            h = module.my_sample(h, (emb, layer_idx))
            hs.append(h)

        if self.double_mid:
                h = th.cat([h, h], dim=1)

        for module in self.output_blocks:
            h = th.cat([h, hs.pop()], dim=1)
            h = module.my_sample(h, (emb, layer_idx))
        #####
        #h = h.type(x.dtype)
        h = h.type(th.cuda.FloatTensor)
        out = self.out.my_sample(h, (None, layer_idx))
        assert pointer.p == len(layer_idx)
        out = out.type(x.dtype)

        return out


class SuperResModel(UNetModel):
    """
    A UNetModel that performs super-resolution.

    Expects an extra kwarg `low_res` to condition on a low-resolution image.
    """

    def __init__(self, image_size, in_channels, *args, **kwargs):
        super().__init__(image_size, in_channels * 2, *args, **kwargs)

    def forward(self, x, timesteps, low_res=None, **kwargs):
        _, _, new_height, new_width = x.shape
        upsampled = F.interpolate(low_res, (new_height, new_width), mode="bilinear")
        x = th.cat([x, upsampled], dim=1)
        return super().forward(x, timesteps, **kwargs)


class EncoderUNetModel(nn.Module):
    """
    The half UNet model with attention and timestep embedding.

    For usage, see UNet.
    """

    def __init__(
        self,
        image_size,
        in_channels,
        model_channels,
        out_channels,
        num_res_blocks,
        attention_resolutions,
        dropout=0,
        channel_mult=(1, 2, 4, 8),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=1,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=False,
        resblock_updown=False,
        use_new_attention_order=False,
        pool="adaptive",
    ):
        super().__init__()

        if num_heads_upsample == -1:
            num_heads_upsample = num_heads

        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.attention_resolutions = attention_resolutions
        self.dropout = dropout
        self.channel_mult = channel_mult
        self.conv_resample = conv_resample
        self.use_checkpoint = use_checkpoint
        self.dtype = th.float16 if use_fp16 else th.float32
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.num_heads_upsample = num_heads_upsample

        time_embed_dim = model_channels * 4
        self.time_embed = nn.Sequential(
            linear(model_channels, time_embed_dim),
            nn.SiLU(),
            linear(time_embed_dim, time_embed_dim),
        )

        ch = int(channel_mult[0] * model_channels)
        self.input_blocks = nn.ModuleList(
            [TimestepEmbedSequential(conv_nd(dims, in_channels, ch, 3, padding=1))]
        )
        self._feature_size = ch
        input_block_chans = [ch]
        ds = 1
        for level, mult in enumerate(channel_mult):
            for _ in range(num_res_blocks):
                layers = [
                    ResBlock(
                        ch,
                        time_embed_dim,
                        dropout,
                        out_channels=int(mult * model_channels),
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                ch = int(mult * model_channels)
                if ds in attention_resolutions:
                    layers.append(
                        AttentionBlock(
                            ch,
                            use_checkpoint=use_checkpoint,
                            num_heads=num_heads,
                            num_head_channels=num_head_channels,
                            use_new_attention_order=use_new_attention_order,
                        )
                    )
                self.input_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch
                input_block_chans.append(ch)
            if level != len(channel_mult) - 1:
                out_ch = ch
                self.input_blocks.append(
                    TimestepEmbedSequential(
                        ResBlock(
                            ch,
                            time_embed_dim,
                            dropout,
                            out_channels=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            down=True,
                        )
                        if resblock_updown
                        else Downsample(
                            ch, conv_resample, dims=dims, out_channels=out_ch
                        )
                    )
                )
                ch = out_ch
                input_block_chans.append(ch)
                ds *= 2
                self._feature_size += ch

        self.middle_block = TimestepEmbedSequential(
            ResBlock(
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
            AttentionBlock(
                ch,
                use_checkpoint=use_checkpoint,
                num_heads=num_heads,
                num_head_channels=num_head_channels,
                use_new_attention_order=use_new_attention_order,
            ),
            ResBlock(
                ch,
                time_embed_dim,
                dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
        )
        self._feature_size += ch
        self.pool = pool
        if pool == "adaptive":
            self.out = nn.Sequential(
                normalization(ch),
                nn.SiLU(),
                nn.AdaptiveAvgPool2d((1, 1)),
                zero_module(conv_nd(dims, ch, out_channels, 1)),
                nn.Flatten(),
            )
        elif pool == "attention":
            assert num_head_channels != -1
            self.out = nn.Sequential(
                normalization(ch),
                nn.SiLU(),
                AttentionPool2d(
                    (image_size // ds), ch, num_head_channels, out_channels
                ),
            )
        elif pool == "spatial":
            self.out = nn.Sequential(
                nn.Linear(self._feature_size, 2048),
                nn.ReLU(),
                nn.Linear(2048, self.out_channels),
            )
        elif pool == "spatial_v2":
            self.out = nn.Sequential(
                nn.Linear(self._feature_size, 2048),
                normalization(2048),
                nn.SiLU(),
                nn.Linear(2048, self.out_channels),
            )
        else:
            raise NotImplementedError(f"Unexpected {pool} pooling")

    def convert_to_fp16(self):
        """
        Convert the torso of the model to float16.
        """
        self.input_blocks.apply(convert_module_to_f16)
        self.middle_block.apply(convert_module_to_f16)

    def convert_to_fp32(self):
        """
        Convert the torso of the model to float32.
        """
        self.input_blocks.apply(convert_module_to_f32)
        self.middle_block.apply(convert_module_to_f32)

    def forward(self, x, timesteps):
        """
        Apply the model to an input batch.

        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :return: an [N x K] Tensor of outputs.
        """
        emb = self.time_embed(timestep_embedding(timesteps, self.model_channels))

        results = []
        h = x.type(self.dtype)
        for module in self.input_blocks:
            h = module(h, emb)
            if self.pool.startswith("spatial"):
                results.append(h.type(x.dtype).mean(dim=(2, 3)))
        h = self.middle_block(h, emb)
        if self.pool.startswith("spatial"):
            results.append(h.type(x.dtype).mean(dim=(2, 3)))
            h = th.cat(results, axis=-1)
            return self.out(h)
        else:
            ####
            #h = h.type(x.dtype)
            h = h.type(self.dtype)
            out = self.out(h)
            out = out.type(self.dtype)
            return out


class Uni_ResBlock(TimestepBlock):
    def __init__(
        self,
        channels,
        emb_channels,
        dropout,
        out_channels=None,
        use_conv=False,
        use_scale_shift_norm=False,
        dims=2,
        use_checkpoint=False,
        up=False,
        down=False,
    ):
        super().__init__()
        self.channels = channels
        self.emb_channels = emb_channels
        self.dropout = dropout
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_checkpoint = use_checkpoint
        self.use_scale_shift_norm = use_scale_shift_norm

        self.in_layers = nn.Sequential(
            # normalization(channels), 将上下采样块之前的东西移到resblock外面
            # nn.SiLU(),将上下采样块之前的东西移到resblock外面
            conv_nd(dims, channels, self.out_channels, 3, padding=1),
        )

        # 将上下采样块移到resblock外面
        # self.updown = up or down
        # if up:
        #     self.h_upd = Upsample(channels, False, dims)
        #     self.x_upd = Upsample(channels, False, dims)
        # elif down:
        #     self.h_upd = Downsample(channels, False, dims)
        #     self.x_upd = Downsample(channels, False, dims)
        # else:
        #     self.h_upd = self.x_upd = nn.Identity()

        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            linear(
                emb_channels,
                2 * self.out_channels if use_scale_shift_norm else self.out_channels,
            ),
        )
        self.out_layers = nn.Sequential(
            normalization(self.out_channels),
            nn.SiLU(),
            nn.Dropout(p=dropout),
            zero_module(
                conv_nd(dims, self.out_channels, self.out_channels, 3, padding=1)
            ),
        )

        # 将skip connection移到resblock外面
        # if self.out_channels == channels:
        #     self.skip_connection = nn.Identity()
        # elif use_conv:
        #     self.skip_connection = conv_nd(
        #         dims, channels, self.out_channels, 3, padding=1
        #     )
        # else:
        #     self.skip_connection = conv_nd(dims, channels, self.out_channels, 1)

    def forward(self, x, emb):
        """
        Apply the block to a Tensor, conditioned on a timestep embedding.

        :param x: an [N x C x ...] Tensor of features.
        :param emb: an [N x emb_channels] Tensor of timestep embeddings.
        :return: an [N x C x ...] Tensor of outputs.
        """
        return checkpoint(
            self._forward, (x, emb), self.parameters(), self.use_checkpoint
        )

    def _forward(self, x, emb):
        # 将上下采样块移到resblock外面
        # if self.updown:
        #     in_rest, in_conv = self.in_layers[:-1], self.in_layers[-1]
        #     h = in_rest(x)
        #     h = self.h_upd(h)
        #     x = self.x_upd(x)
        #     h = in_conv(h)
        # else:
        #     h = self.in_layers(x)
        h = self.in_layers(x)
        emb_out = self.emb_layers(emb).type(h.dtype)
        while len(emb_out.shape) < len(h.shape):
            emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            out_norm, out_rest = self.out_layers[0], self.out_layers[1:]
            scale, shift = th.chunk(emb_out, 2, dim=1)
            h = out_norm(h) * (1 + scale) + shift
            h = out_rest(h)
        else:
            h = h + emb_out
            h = self.out_layers(h)
        # return self.skip_connection(x) + h
        return h


class FNN(nn.Module):
  """Fully connected feedforward neural network."""

  def __init__(self, dims, threshold):
    super(FNN, self).__init__()
    assert len(dims) >= 2, 'Length of dims is smaller than 2.'
    self.threshold = threshold

    self.fc = []
    for i in range(len(dims) - 2):
      self.fc.append(nn.Linear(dims[i], dims[i+1]))
      self.fc.append(nn.ReLU(inplace=True))
    self.fc.append(nn.Linear(dims[-2], dims[-1]))

    self.fc = nn.Sequential(*self.fc)

  def forward(self, x):
    preact = self.fc(x)
    if self.training:
        preact += th.randn_like(preact)
    gate = F.sigmoid(preact)
    gate = th.clamp(1.2 * gate - 0.1, min = 0, max = 1) + 1 # Just use Simultaneous Expansion in DPDM v1.
    discrete_gate = F.sigmoid(100 * (gate - self.threshold))

    if self.training:
        discrete_prob = 0.5
        mix_mask = gate.new_empty(
            size=[gate.size()[0], 1]).uniform_() < discrete_prob
        gate = th.where(mix_mask, discrete_gate, gate)
    else:
        gate = (gate > self.threshold).int()

    return gate


class BigResBlock(TimestepBlock):
    def __init__(
        self,
        expand_scale,
        channels,
        emb_channels,
        dropout,
        out_channels=None,
        use_conv=False,
        use_scale_shift_norm=False,
        dims=2,
        use_checkpoint=False,
        up=False,
        down=False,
    ):
        super().__init__()
        self.unit_res_blocks = nn.ModuleList([])
        self.out_channels = out_channels or channels
        self.expand_scale = expand_scale
        self.use_checkpoint = use_checkpoint

        for i in range(expand_scale):
            self.unit_res_blocks.append(
                Uni_ResBlock(
                    channels,
                    emb_channels,
                    dropout,
                    out_channels,
                    use_conv,
                    use_scale_shift_norm,
                    dims,
                    use_checkpoint,
                    up,
                    down,
                )
            )
        
        # 将上下采样块移到resblock外面
        self.layers_before_upd = nn.Sequential(
            normalization(channels),
            nn.SiLU()
        )
        self.updown = up or down
        if up:
            self.h_upd = Upsample(channels, False, dims)
            self.x_upd = Upsample(channels, False, dims)
        elif down:
            self.h_upd = Downsample(channels, False, dims)
            self.x_upd = Downsample(channels, False, dims)
        else:
            self.h_upd = self.x_upd = nn.Identity()


        # 将skip connection移到resblock外面
        if self.out_channels == channels:
            self.skip_connection = nn.Identity()
        elif use_conv:
            self.skip_connection = conv_nd(
                dims, channels, self.out_channels, 3, padding=1
            )
        else:
            self.skip_connection = conv_nd(dims, channels, self.out_channels, 1)

    # def forward(self, x, emb):
    #     assert len(emb) == 2
    #     return checkpoint(
    #         self._forward, (x, emb), self.parameters(), self.use_checkpoint
    #     )

    def forward(self, x, emb):
        assert len(emb) == 2
        start_id = pointer.p
        end_id = start_id + self.expand_scale
        gate = emb[1][:, start_id : end_id].unsqueeze(dim=2).unsqueeze(dim=3).unsqueeze(dim=4)
        pointer.p = end_id

        hx = self.layers_before_upd(x)
        if self.updown:
            x = self.x_upd(x)
            hx = self.h_upd(hx)

        h = gate[:, 0] * self.unit_res_blocks[0](hx, emb[0])
        for i in range(self.expand_scale-1):
            idx = i + 1
            h += gate[:, idx] * self.unit_res_blocks[idx](hx, emb[0])

        sum_gate = th.sum(gate + 1e-5, dim=1)
        return self.skip_connection(x) + h / sum_gate

    # def my_sample(self, x, emb):
    #     assert len(emb) == 2

    #     return checkpoint(
    #         self.my_sample_, (x, emb), self.parameters(), self.use_checkpoint
    #     )

    def my_sample(self, x, emb):
        assert len(emb) == 2
        layer_idx = emb[1][pointer.p]
        pointer.p += 1
        hx = self.layers_before_upd(x)
        if self.updown:
            x = self.x_upd(x)
            hx = self.h_upd(hx)

        if len(layer_idx) > 0:
            '''num = -1
            for i in layer_idx:
                num += 1
                if num == 0:
                    h = self.unit_res_blocks[i](hx, emb[0])
                else:
                    h += self.unit_res_blocks[i](hx, emb[0])'''
            h = self.unit_res_blocks[layer_idx[0]](hx, emb[0])

            return self.skip_connection(x) + h
            # return self.skip_connection(x) + h / float(len(layer_idx))

        else:
            return self.skip_connection(x)


class Input0(TimestepBlock):
    def __init__(
        self,
        expand_scale,
        dims,
        in_channels,
        out_channels,
    ):
        super().__init__()
        self.unit_conv_blocks = nn.ModuleList([])
        self.expand_scale = expand_scale

        for i in range(expand_scale):
            self.unit_conv_blocks.append(
                conv_nd(dims, in_channels, out_channels, 3, padding=1)
            )

    def forward(self, x, emb):
        assert len(emb) == 2
        start_id = pointer.p
        end_id = start_id + self.expand_scale
        gate = emb[1][:, start_id : end_id].unsqueeze(dim=2).unsqueeze(dim=3).unsqueeze(dim=4)
        pointer.p = end_id

        h = gate[:, 0] * self.unit_conv_blocks[0](x)
        for i in range(self.expand_scale-1):
            idx = i + 1
            h += gate[:, idx] * self.unit_conv_blocks[idx](x)

        sum_gate = th.sum(gate + 1e-5, dim=1)
        return h / sum_gate

    def my_sample(self, x, emb):
        assert len(emb) == 2
        layer_idx = emb[1][pointer.p]
        pointer.p += 1

        assert len(layer_idx) > 0
        '''num = -1
        for i in layer_idx:
            num += 1
            if num == 0:
                h = self.unit_conv_blocks[i](x)
            else:
                h += self.unit_conv_blocks[i](x)'''
        h = self.unit_conv_blocks[layer_idx[0]](x)

        return h


class Out_layer(nn.Module):
    def __init__(
        self,
        expand_scale,
        dims,
        ch,
        in_channels,
        out_channels,
    ):
        super().__init__()
        self.unit_conv_blocks = nn.ModuleList([])
        self.expand_scale = expand_scale

        for i in range(expand_scale):
            self.unit_conv_blocks.append(
                nn.Sequential(
                    normalization(ch),
                    nn.SiLU(),
                    zero_module(conv_nd(dims, in_channels, out_channels, 3, padding=1)),
                )
            )

    def forward(self, x, emb):
        assert len(emb) == 2
        start_id = pointer.p
        end_id = start_id + self.expand_scale
        gate = emb[1][:, start_id : end_id].unsqueeze(dim=2).unsqueeze(dim=3).unsqueeze(dim=4)
        pointer.p = end_id

        h = gate[:, 0] * self.unit_conv_blocks[0](x)
        for i in range(self.expand_scale-1):
            idx = i + 1
            h += gate[:, idx] * self.unit_conv_blocks[idx](x)

        # sum_gate = th.mean(gate.squeeze(dim=4).squeeze(dim=3).squeeze(dim=2), dim=1).unsqueeze(dim=1).unsqueeze(dim=2).unsqueeze(dim=3)
        sum_gate = th.sum(gate + 1e-10, dim=1)
        return h / sum_gate

    def my_sample(self, x, emb):
        assert len(emb) == 2
        layer_idx = emb[1][pointer.p]
        pointer.p += 1

        assert len(layer_idx) > 0
        '''num = -1
        for i in layer_idx:
            num += 1
            if num == 0:
                h = self.unit_conv_blocks[i](x)
            else:
                h += self.unit_conv_blocks[i](x)'''
        h = self.unit_conv_blocks[layer_idx[0]](x)

        return h


class UniAttentionBlock(nn.Module):
    def __init__(
        self,
        channels,
        num_heads=1,
        num_head_channels=-1,
        use_checkpoint=False,
        use_new_attention_order=False,
    ):
        super().__init__()
        self.channels = channels
        if num_head_channels == -1:
            self.num_heads = num_heads
        else:
            assert (
                channels % num_head_channels == 0
            ), f"q,k,v channels {channels} is not divisible by num_head_channels {num_head_channels}"
            self.num_heads = channels // num_head_channels
        self.use_checkpoint = use_checkpoint
        self.norm = normalization(channels)
        self.qkv = conv_nd(1, channels, channels * 3, 1)
        if use_new_attention_order:
            # split qkv before split heads
            self.attention = QKVAttention(self.num_heads)
        else:
            # split heads before split qkv
            self.attention = QKVAttentionLegacy(self.num_heads)

        self.proj_out = zero_module(conv_nd(1, channels, channels, 1))

    def forward(self, x):
        return checkpoint(self._forward, (x,), self.parameters(), True)

    def _forward(self, x):
        b, c, *spatial = x.shape
        # x = x.reshape(b, c, -1)
        qkv = self.qkv(self.norm(x))
        h = self.attention(qkv)
        h = self.proj_out(h)
        # return (x + h).reshape(b, c, *spatial)
        return h


class BigAttentionBlock(TimestepBlock):
    def __init__(
        self,
        expand_scale,
        channels,
        num_heads=1,
        num_head_channels=-1,
        use_checkpoint=False,
        use_new_attention_order=False,
    ):
        super().__init__()
        self.att_blocks = nn.ModuleList([])
        self.expand_scale = expand_scale

        for i in range(expand_scale):
            self.att_blocks.append(
                UniAttentionBlock(
                    channels,
                    num_heads,
                    num_head_channels,
                    use_checkpoint,
                    use_new_attention_order,
                )
            )

    def forward(self, x, emb):
        b, c, *spatial = x.shape
        x = x.reshape(b, c, -1)
        assert len(emb) == 2
        start_id = pointer.p
        end_id = start_id + self.expand_scale
        gate = emb[1][:, start_id : end_id].unsqueeze(dim=2).unsqueeze(dim=3)
        pointer.p = end_id

        h = gate[:, 0] * self.att_blocks[0](x)
        for i in range(self.expand_scale-1):
            idx = i + 1
            h += gate[:, idx] * self.att_blocks[idx](x)

        sum_gate = th.sum(gate + 1e-5, dim=1)
        return (x + h / sum_gate).reshape(b, c, *spatial)

    def my_sample(self, x, emb):
        assert len(emb) == 2

        layer_idx = emb[1][pointer.p]
        pointer.p += 1
        b, c, *spatial = x.shape
        x = x.reshape(b, c, -1)

        if len(layer_idx) > 0:
            h = self.att_blocks[layer_idx[0]](x)

            return (x + h).reshape(b, c, *spatial)

        else:
            return x.reshape(b, c, *spatial)